use std::fmt::Display;
use termion::color;

use crate::{Player, Symb};

#[derive(Debug, PartialEq)]
enum Token {
	Number(f32),
	OpAdd,
	OpSub,
	OpMult,
	OpDiv,
}

impl Token {
	fn val(&self) -> f32 {
		match self {
			Self::Number(x) => *x,
			_ => unreachable!(),
		}
	}
}

pub struct Board {
	board: [Option<(Symb, Player)>; 11],
	free_spots: usize,
	current_player: Player,
}

impl Display for Board {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		// Print board
		for i in &self.board {
			match i {
				Some((symb, player)) => write!(
					f,
					"{}{}{}",
					color::Fg(player.color()),
					symb,
					color::Fg(color::Reset)
				)?,

				None => write!(f, "_")?,
			}
		}
		Ok(())
	}
}

impl Board {
	pub fn new(current_player: Player) -> Self {
		Self {
			free_spots: 11,
			board: Default::default(),
			current_player,
		}
	}

	pub fn current_player(&self) -> Player {
		self.current_player
	}

	pub fn is_done(&self) -> bool {
		self.free_spots == 0
	}

	pub fn size(&self) -> usize {
		self.board.len()
	}

	// Place the marked symbol at the given position.
	// Returns true for valid moves and false otherwise.
	pub fn play(&mut self, cursor: usize, symb: Symb) -> bool {
		match &self.board[cursor] {
			Some(_) => return false,
			None => {
				// Check for duplicate symbols
				for i in &self.board {
					if let Some(i) = i {
						if i.0 == symb {
							return false;
						}
					}
				}

				// Check syntax
				match symb {
					Symb::Minus => {
						if cursor == self.board.len() - 1 {
							return false;
						}

						let r = &self.board[cursor + 1];
						if r.is_some_and(|(s, _)| s.is_binop()) {
							return false;
						}
					}

					Symb::Zero => {
						if cursor != 0 {
							let l = &self.board[cursor - 1].map(|x| x.0);
							if l == &Some(Symb::Div) {
								return false;
							}
						}
					}

					Symb::Div | Symb::Plus | Symb::Times => {
						if cursor == 0 || cursor == self.board.len() - 1 {
							return false;
						}

						let l = &self.board[cursor - 1].map(|x| x.0);
						let r = &self.board[cursor + 1].map(|x| x.0);

						if symb == Symb::Div && r == &Some(Symb::Zero) {
							return false;
						}

						if l.is_some_and(|s| s.is_binop() || s.is_minus())
							|| r.is_some_and(|s| s.is_binop())
						{
							return false;
						}
					}
					_ => {}
				}

				self.board[cursor] = Some((symb, self.current_player));
			}
		}

		self.current_player.invert();
		self.free_spots -= 1;
		true
	}

	pub fn evaluate(&self) -> Option<f32> {
		if !self.is_done() {
			return None;
		}

		let mut tokens = Vec::new();
		let mut is_neg = true; // if true, - is negative. if false, subtract.
		let mut current_num = 0f32;
		let mut current_sgn = 1f32;

		for (s, _) in self.board.iter().map(|x| x.unwrap()) {
			match s {
				Symb::Div => {
					tokens.push(Token::Number(current_num * current_sgn));
					current_num = 0.0;
					current_sgn = 1.0;
					tokens.push(Token::OpDiv);
					is_neg = true;
				}
				Symb::Minus => {
					if is_neg {
						current_sgn = -1.0;
					} else {
						tokens.push(Token::Number(current_num * current_sgn));
						current_num = 0.0;
						current_sgn = 1.0;
						tokens.push(Token::OpSub);
						is_neg = true;
					}
				}
				Symb::Plus => {
					tokens.push(Token::Number(current_num * current_sgn));
					current_num = 0.0;
					current_sgn = 1.0;
					tokens.push(Token::OpAdd);
					is_neg = true;
				}
				Symb::Times => {
					tokens.push(Token::Number(current_num * current_sgn));
					current_num = 0.0;
					current_sgn = 1.0;
					tokens.push(Token::OpMult);
					is_neg = true;
				}
				Symb::Zero => {
					current_num = current_num * 10.0;
					is_neg = false;
				}
				Symb::Number(x) => {
					current_num = current_num * 10.0 + (x.get() as f32);
					is_neg = false;
				}
			}
		}

		tokens.push(Token::Number(current_num * current_sgn));

		for op in [Token::OpMult, Token::OpDiv, Token::OpAdd, Token::OpSub] {
			for i in 0..tokens.len() {
				if tokens[i] == op {
					let l = &tokens[i - 1];
					let r = &tokens[i + 1];

					let v = match op {
						Token::OpAdd => l.val() + r.val(),
						Token::OpDiv => l.val() / r.val(),
						Token::OpSub => l.val() - r.val(),
						Token::OpMult => l.val() * r.val(),
						_ => unreachable!(),
					};

					tokens.remove(i - 1);
					tokens.remove(i - 1);
					tokens[i - 1] = Token::Number(v);
					break;
				}
			}
		}

		return Some(tokens[0].val());
	}
}
